import Output as _Output
import warnings

def MakeTrackFiles(savelineFileName, line, outputFileNameStub):
    sl  = _Output.Saveline(savelineFileName, line)
    sl.removeDuplicates()
    sl.writeRenamed(outputFileNameStub+"_renamed.mad8")

    observeElements = []
    observeIndex    = [] 
    for e in sl.expandedLineRenamed : 
        d = sl.findRenamedNamedDict(e) 
        if d['type'] != 'DRIFT' : 
            observeElements.append(e)
            observeIndex.append(sl.expandedLineRenamed.index(e))

    MakeObserveFile(observeElements,outputFileNameStub+"_observe.mad8")
    MakeTableArchiveFile(observeElements,outputFileNameStub+"_archive.mad8")
    MakeTableMapFile(observeElements,observeIndex,outputFileNameStub+"_trackFileMap.mad8")
    MakeInraysFile(bdsimOutput, outputFileNameStub+"_inrays.mad8")
    MakeTrackCallingFile(outputFileNameStub)

def MakeTrackCallingFile(fileNameStub) : 
    pass

def MakeTableMapFile(observeElements, observeIndex, filename) :
    f = open(filename,'w')
    
    for i in range(0,len(observeElements)) : 
        f.write(str(observeIndex[i])+' '+observeElements[i]+'\n')

    f.close()

def MakeObserveFile(elementlist,filename) : 
    f = open(filename,'w') 
    for e in elementlist : 
        ws = 'OBSERVE, PLACE="'+e+'", TABLE="'+e+'"\n'
        f.write(ws)
    f.close()

def MakeTableArchiveFile(elementlist, filename) :
    f = open(filename,'w')
    for e in elementlist :
        ws = 'ARCHIVE, TABLE="'+e+'", FILENAME="./track/'+e+'"\n'
        f.write(ws)
    f.close()



import pylab as pl 
import numpy as np 
import Mad8  as m8

def testTrack(rmatFile, nparticle = 10) :
    o = m8.OutputReader()     
    [c,r] = o.readFile(rmatFile,"rmat") 
    c.subline('IEX',-1) 
    r.subline('IEX',-1)
    s = Track(c,r)
    s.trackParticles(nparticle)
    return s

class Track : 
    def __init__(self, common, rmat) : 
        self.common     = common
        self.rmat       = rmat
        self.nelement   = rmat.data.shape[0]
        # Turn rmat data into matrix
        self.rmatMatrix = np.reshape(rmat.data[:,0:36],(self.nelement,6,6))

    def trackParticles(self, nparticle) : 
        # store the number of particles tracked 
        self.nparticle = nparticle
        
        # Nparticle, Nelement phase space storage
        self.psVector = np.zeros((nparticle,self.nelement,6))

        # loop over particles
        for i in range(0,nparticle) : 
            print 'Track.Simple.track> particle ',i
            p = self.generate() 
            psv = self.trackParticle(p)
            self.psVector[i] = psv
    
    def generate(self) : 
        return np.array([0.0,
                         0.0,
                         pl.normal(0,1e-6),
                         0.0,
                         0.0,
                         0.0],dtype=np.double)

    def trackParticle(self,p) :

        # Tracking vector 
        psVector   = np.zeros((self.rmatMatrix.shape[0],6))

        # Set first element
        i = 1 
        psVector[0] = p 
                
        # loop over all elements from start to end 
        for m in self.rmatMatrix[1:] :
#            psVector[i] = np.dot(m,psVector[i-1])
            psVector[i] = np.dot(m,p)
            i=i+1

        return psVector




"""
pymad8 - python tools for working with MAD8 output and input.

| Dependencies:
| package     - minimum version required
| numpy       - 1.7.1
| matplotlib  - 1.3.0

| Modules:
| Input         -
| Output        -
| Plot          -
| Sim           -
| Track         -
| Visualisation -

Copyright Royal Holloway, University of London 2018.
"""

__version__ = "1.3"

# import Mad8
import Input
import Output
import Plot
import Sim
#import Track  #not imported by default - can be explicitly imported
import Visualisation

#import Saveline
#import Converter
#import Track

__all__ = ['Input',
           'Output',
           'Plot',
           'Sim',
           'Visualisation'
]



import pylab as pl 
import pymad8 as m8 
import matplotlib as mpl
import matplotlib.patches as mpt
import matplotlib.pyplot  as plt

def testOneDim() : 
    o = m8.Output.OutputReader()
    [c,s] = o.readFile("./test/atf_v5.1/survey.tape","survey")
    od = OneDim(c,s,False)
    od.plot()
    return od    

def testTwoDim() : 
    o = m8.Output.OutputReader()
    [c,s] = o.readFile("./test/atf_v5.1/survey.tape","survey")
    td = TwoDim(c,s,False,False,True)
    td.plot()
    return td

def transformedRect(xyc, dx, dy, theta) : 
    x = xyc[0]
    y = xyc[1]

    # basic rectangle
    xy = pl.array([[-dx/2.0, -dy/2.0],
                   [-dx/2.0, +dy/2.0],
                   [+dx/2.0, +dy/2.0],
                   [+dx/2.0, -dy/2.0]])

    # transform to correct location 
    p  = transformedPoly(xy,xyc,theta)
    
    return p

def transformedPoly(xy, xyc, theta) :
    # Rotate in place
    r = pl.array([[ pl.cos(theta),-pl.sin(theta)],
                  [ pl.sin(theta), pl.cos(theta)]])
    xy = xy.dot(r)

    # Translate to new centre 
    xy = xy + xyc 
    p = mpt.Polygon(xy)

    # Return transformed poly
    return p

class OneDim :
    def __init__(self, common, survey, debug) :
        self.common = common
        self.survey = survey 

        self.x    =  self.survey.data[:,self.survey.keys['x']]
        self.y    =  self.survey.data[:,self.survey.keys['y']]
        self.z    = -self.survey.data[:,self.survey.keys['z']]
        self.suml =  self.survey.data[:,self.survey.keys['suml']]

        self.debug    = debug 
        #self.annotate = annotate 
        self.quadWidth = 0.1
        self.bendWidth = 0.1
        self.sextWidth = 0.1

        self._offcolour = '0.9'
        self._no_colour = '0.2'

    def plot(self,colour=True) : 
        s = self.survey.data[:,self.survey.keys['suml']]
        z = pl.zeros(s.shape)

        # plot beam line
        pl.plot(s,z,'k-')
        pl.ylim(-1.5,1.5)

        # Draw specific types of element
        self.drawElements("QUAD",colour) 
        self.drawElements("SBEN",colour) 
        self.drawElements("RBEN",colour)
        self.drawElements("SEXT",colour)

    def drawElements(self,type,colour=True) : 
        if self.debug : 
            print 'pymad8.Visualisation.OneDim.drawElements>'

        ilist = self.common.findByType(type) 
        for i in ilist : 
            self.drawElement(i,colour)
    
    def drawElement(self,elem,colour=True) : 
        if self.debug : 
            print 'pymad8.Visualisation.OneDim.drawElement>',elem
            print 'pymad8.Visualisation.OneDim.drawElement> Use Colour -> ',colour

        # find element if string
        if type(elem) == str : 
            i = self.common.findByName(name)[0]
        else :
            i = elem

        t    = self.common.type[i] 
        n    = self.common.name[i]
        c    = self.common.data[i]
        s    = self.survey.data[i]
        suml = self.suml[i]

        if t == 'QUAD' : 
            self.drawQuad(c,s,suml,colour)
        elif t == 'MULT' : 
            self.drawMult(c,s,suml,colour)
        elif t == 'SBEN' : 
            self.drawBend(c,s,suml,colour)
        elif t == 'RBEN' : 
            self.drawBend(c,s,suml,colour)
        elif t == 'SEXT' : 
            self.drawSext(c,s,suml,colour)
        elif t == 'HKIC' :
            self.drawHkic(c,s,suml,colour)
        elif t == 'VKIC' : 
            self.drawVkic(c,s,suml,colour)
        elif t == 'MONI' : 
            self.drawMoni(c,s,suml,colour)
        elif t == 'WIRE' : 
            self.drawWire(c,s,suml,colour)
        elif t == 'PROF' : 
            self.drawProf(c,s,suml,colour)
        elif t == 'INST' : 
            self.drawInst(c,s,suml,colour)
        elif t == 'MARK' : 
            self.drawMark(c,s,suml,colour)
        else :
            print 'pymad8.Visualisation.OneDim> Type not known'

        # Annotate element

    def drawQuad(self,c,s,suml,colour=True) : 
        ql = c[m8.Output.Common.keys['quad']['l']]
        qk = c[m8.Output.Common.keys['quad']['k1']]
        poscolour = 'r'
        negcolour = 'b'
        if colour == False:
                poscolour,negcolour = self._no_colour,self._no_colour
        
        if qk > 0 :
            qr = mpt.Rectangle((suml-ql/2.0,0),
                               ql,self.quadWidth,color=poscolour)
        elif qk < 0 : 
            qr = mpt.Rectangle((suml-ql/2.0,-self.quadWidth),
                               ql,self.quadWidth,color=negcolour)            
        elif qk == 0 : 
            qr = mpt.Rectangle((suml-ql/2.0,-self.quadWidth/2.0),
                               ql,self.quadWidth,color=self._offcolour)            
            
        ax = plt.gca()            
        ax.add_patch(qr)

    def drawMult(self,c,s,suml,colour=True) :
        pass

    def drawBend(self,c,s,suml,colour=True) : 
        #colour argument unused but added for compliance
        bl = c[m8.Output.Common.keys['sben']['l']]
        br = mpt.Rectangle((suml-bl/2.0,-self.bendWidth/2.0),
                           bl,self.bendWidth,color='k')
        ax = plt.gca()            
        ax.add_patch(br)

    def drawSext(self,c,s,suml,colour=True) : 
        sl = c[m8.Output.Common.keys['sextupole']['l']]
        ax = plt.gca()
        #original but looks like thin lens for most lattices rather
        #than hexagon
        #sh = mpt.RegularPolygon((suml-sl/2.0,0),6,
        #                        sl,color='g')
        if colour == True:
            sext_colour = 'g'
        else:
            sext_colour = 'grey'
        
        sh = mpt.Rectangle((suml-sl/2.0,-self.bendWidth/2.0),
                           sl,self.bendWidth,color=sext_colour) 
        ax.add_patch(sh)        

    def drawHkic(self,c,s,suml,colour=True) :
        pass

    def drawVkic(self,c,s,suml,colour=True) :
        pass
        
    def drawMoni(self,c,s,suml,colour=True) :
        pass
    
    def drawWire(self,c,s,suml,colour=True) :
        pass

    def drawProf(self,c,s,suml,colour=True) :
        pass

    def drawInst(self,c,s,suml,colour=True) :
        pass
    
    def drawMark(self,c,s,suml,colour=True) :
        pass


class TwoDim : 
    def __init__(self,common,survey, debug = False, annotate = False, fancy = False) :
        self.debug    = debug
        self.annotate = annotate
        self.fancy    = fancy
        self.common   = common
        self.survey   = survey
        self.quadWidth= 0.75

    def plot(self,event=None) : 
        print 'Visualisation.TwoDim.plot>'
        self.x = self.survey.data[:,self.survey.keys['x']]
        self.y = self.survey.data[:,self.survey.keys['y']]
        self.z = -self.survey.data[:,self.survey.keys['z']]

        self.f  = plt.figure()
#        self.rec = self.f.canvas.mpl_connect('draw_event',self.plotUpdate)        

        xmin = self.x.min()
        xmax = self.x.max()
        zmin = self.z.min()
        zmax = self.z.max()

        self.ax = plt.gca()
        self.ax.clear()        

        pl.plot(self.z,self.x,'--')        
        pl.xlim(zmin-10,zmax+10)
        pl.ylim(xmin-10,xmax+10)        

        self.drawElements("QUAD") 

    def plotUpdate(self,event) : 
        print 'Visualisation.TwoDim.plotUpdate>'
        self.drawElements("QUAD")

    def drawElements(self,type) : 
        if self.debug : 
            print 'pymad8.Visualisation.TwoDim.drawElements>'

        ilist = self.common.findByType(type) 
        for i in ilist : 
            self.drawElement(i)

    def drawElement(self,elem) : 
        if self.debug : 
            print 'pymad8.Visualisation.TwoDim.drawElement>',elem

        # find element if string
        if type(elem) == str : 
            i = self.common.findByName(name)[0]
        else :
            i = elem

        t = self.common.type[i] 
        n = self.common.name[i]
        c = self.common.data[i]
        s = self.survey.data[i]

        # plot marker
        ex = self.x[i]
        ey = self.y[i]
        ez = self.z[i]

        if t == 'QUAD' : 
            self.drawQuad(c,s,ex,ey,ez)
        elif t == 'SBEN' : 
            self.drawBend(c,s,ex,ey,ez) 
        elif t == 'RBEN' : 
            self.drawBend(c,s,ex,ey,ez)
        elif t == 'MONI' : 
            self.drawMoni(c,s,ex,ey,ez)            
        elif t == 'MARK' : 
            self.drawMark(c,s,ex,ey,ez)
        else :
            print 'pymad8.Visualisation.TwoDim> Type not known'
        # pl.plot([z[i]],[x[i]],"+")

        # Annotate element

    def drawQuad(self,c,s,x,y,z) : 
        if self.debug : 
            print 'Visualisation.TwoDim.drawQuad>'
            print '>',c
            print '>',s
        if self.fancy : 
            # get data
            ql = c[m8.Output.Common.keys['quad']['l']]
            qk = c[m8.Output.Common.keys['quad']['k1']]
            qt = s[m8.Output.Survey.keys['theta']]

            # make patch
            qr = transformedRect([z,x],ql,self.quadWidth,qt)
            qr.set_color('r')
            qr.set_alpha(0.6)

            # add patch
            ax = plt.gca()
            ax.add_patch(qr)
        else : 
            pl.plot([z],[x],'r+')

    def drawBend(self,c,s,x,y,z) : 
        if self.debug : 
            print 'Visualisation.TwoDim.drawDipole>'

        if self.fancy : 
            # get data
            bl = c[m8.Output.Common.keys['sben']['l']]
            bt = s[m8.Output.Survey.keys['theta']]
            
        else : 
            pl.plot([z],[x],'b+')        

    def drawMoni(self,c,s,x,y,z) :
        if self.debug : 
            print 'Visualisation.TwoDim.drawMoni>'

        pl.plot([z],[x],'g+')                

    def drawMark(self,c,s,x,y,z) : 
        if self.debug : 
            print 'Visualisation.TwoDim.drawMark>'

        pl.plot([z],[x],'b+')        
    



from setuptools import setup, find_packages

setup(
    name='pymad8',
    version='1.3',
    packages=find_packages(exclude=["docs", "tests", "old"]),
    # Not sure how strict these need to be...
    install_requires=["matplotlib >= 1.7.1",
                      "numpy >= 1.3.0",
                      "fortranformat >= 0.2.5"],
    # Some version of python2.7
    python_requires="==2.7.*",

    author='JAI@RHUL',
    author_email='stewart.boogert@rhul.ac.uk',
    description="Write MAD8 models and load MAD8 output.",
    url='https://bitbucket.org/jairhul/pymad8/',
    license='GPL3',
    keywords='mad8 accelerator twiss'
)



import pylab as _pl 
import numpy as _np
import sys as _sys
import copy as _copy
import os as _os
from   collections import defaultdict
import pymad8 as _pymad8
import fortranformat as _ff

import Input as _Input

###############################################################################################################
def getValueByName(name, key, common, table) : 
    ind1 = common.findByName(name)
    ind2 = table.keys[key]
    return table.data[ind1,ind2]
###############################################################################################################
class General : 
    '''General list of accelerator component infomation'''
    def __init__(self) : 
        '''No arguments, sets up internal variables for general tables 
        type     : type of element
        name     : name of element
        dataList : parameters associated with element'''

        self.type     = [] 
        self.name     = []
        self.dataList = [] 
        self.ind      = []

    def addElement(self,type,name,data) :
        # just in case strip the type and name of trailing spaces
        type = type.strip(' ')
        name = name.strip(' ')

        if type == '' :
            type == 'init'

        self.type.append(type) 
        self.name.append(name) 
        self.dataList.append(data)

    def makeArray(self) :
        self.data = _np.array(self.dataList)
        self.ind  = _np.arange(0,len(self.data),1)

    def getIndex(self,name) : 
        ind = self.ind[_np.array(self.name) == name]
        return ind
        
    def findByName(self,name) :
        ind = self.getIndex(name)
        if len(ind) == 1 : 
            return ind[0]
        else : 
            return ind

    def findByType(self,type) :
        return self.ind[_np.array(self.type) == type]

    def getNElements(self) : 
        return len(self.name)

    def getNames(self, ind) : 
        return _np.array(self.name)[ind]

    def getRowByName(self, name) : 
        return self.getRowByIndex(self.findByName(name))

    def getRowByIndex(self, index) : # was getData
        d = {}
        d['name'] = self.name[index].strip().lower()
        d['type'] = self.type[index].strip().lower()
        dKeys = self.keys
        
        for k in dKeys :             
            d[k] = self.data[index,dKeys[k]]

        return d

    def getColumn(self, key) : 
        return self.data[:,self.keys[key]]

    def subline(self, start, end) : 
        if type(start) == str : 
            start = self.findByName(start) 
        if type(end) == str : 
            end = self.findByName(end)

        self.type = self.type[start:end]
        self.name = self.name[start:end]
        self.data = self.data[start:end]
        self.ind  = _pl.arange(0,len(self.data),1)
#        self.ind  = self.ind[start:end]


    def plotXY(self, xkey, ykey) :        
        _pl.plot(self.data[:,self.keys[xkey]],
                self.data[:,self.keys[ykey]])

###############################################################################################################
class Common(General) : 

    keys = {
        'drif'       :{'l':0                                                                              ,'aper':9,'note':10,'E':11},
        'rben'       :{'l':0, 'angle':1, 'k1':2 , 'k2':3, 'tilt':4 ,'e1':5    , 'e2':6   , 'h1':7 , 'h2':8,'aper':9,'note':10,'E':11},
        'sben'       :{'l':0, 'angle':1, 'k1':2 , 'k2':3, 'tilt':4 ,'e1':5    , 'e2':6   , 'h1':7 , 'h2':8,'aper':9,'note':10,'E':11},        
        'quad'       :{'l':0,            'k1':2 ,         'tilt':4                                        ,'aper':9,'note':10,'E':11},
        'sext'       :{'l':0,                    'k2':3 , 'tilt':4                                        ,'aper':9,'note':10,'E':11},
        'octu'       :{'l':0,                             'tilt':4 ,'k3':5                                ,'aper':9,'note':10,'E':11},     
        'mult'       :{       'k0l':1  , 'k1l':2,'k2l':3, 't0':4   ,'k3l':5   , 't1':6   , 't2':7 , 't3':8,'aper':9,'note':10,'E':11},
        'sole'       :{'l':0,                                       'ks':5                                ,'aper':9,'note':10,'E':11},
        'rfcavity'   :{'l':0,                                       'freq':5  , 'volt':6 , 'lag':7        ,'aper':9,'note':10,'E':11},
        'elseparator':{'l':0,                             'tilt':4 ,'efield':5                            ,'aper':9,'note':10,'E':11},
        'kick'       :{'l':0,                             'hkick':4,'vkick':5                             ,'aper':9,'note':10,'E':11},    
        'hkic'       :{'l':0,                             'hkick':4                                       ,'aper':9,'note':10,'E':11},    
        'vkic'       :{'l':0,                                       'vkick':5                             ,'aper':9,'note':10,'E':11}, 
        'srot'       :{'l':0,                                       'angle':5                             ,'aper':9,'note':10,'E':11}, 
        'yrot'       :{'l':0,                                       'angle':5                             ,'aper':9,'note':10,'E':11}, 
        'moni'       :{'l':0                                                                              ,'aper':9,'note':10,'E':11},
        'hmonitor'   :{'l':0                                                                              ,'aper':9,'note':10,'E':11},
        'vmonitor'   :{'l':0                                                                              ,'aper':9,'note':10,'E':11},
        'mark'       :{'l':0                                                                              ,'aper':9,'note':10,'E':11},
        'ecol'       :{'l':0,                             'xsize':4,'ysize':5                             ,'aper':9,'note':10,'E':11},   
        'rcol'       :{'l':0,                             'xsize':4,'ysize':5                             ,'aper':9,'note':10,'E':11},
        'mark'       :{'l':0,                                                                                       'note':10,'E':11},
        'inst'       :{'l':0,                                                                                       'note':10,'E':11},
        'wire'       :{'l':0,                                                                                       'note':10,'E':11},
        'imon'       :{'l':0,                                                                                       'note':10,'E':11},
        'prof'       :{'l':0,                                                                                       'note':10,'E':11},
        'blmo'       :{'l':0,                                                                                       'note':10,'E':11},
        'lcav'       :{'l':0,                                       'freq':5  , 'volt':6 , 'lag':7        ,'aper':9,'note':10,'E':11},
        'matr'       :{'l':0,                                                                              'aper':9,          'E':11}
    }

    def __init__(self) :
        General.__init__(self)

    def getData(self, index) :
        d = {}
        d['name'] = self.name[index].strip().lower()
        d['type'] = self.type[index].strip().lower()
        dKeys = self.keys[self.type[index].lower()]
        
        for k in dKeys : 
            d[k] = self.data[index,dKeys[k]]

        return d

    def getRowByName(self, name) : 
        return self.getRowByIndex(self.findByName(name))

    def getRowByIndex(self, index) : # was getData
        d = {}
        d['name'] = self.name[index].strip().lower()
        d['type'] = self.type[index].strip().lower()
        dKeys = self.keys

        if d['type'] == '' : 
            return d

        for k in dKeys[d['type']] :
            d[k] = self.data[index,dKeys[d['type']][k]]
        
        try : 
            self.keys.keys().index(d['type'])
            for k in dKeys[d['type']] : 
                if k != 'note' : 
                    d[k] = float(d[k])
            return d
        except ValueError : 
            return d


    def getColumn(self,colName) : 
        d  = [] 
        if colName == "E" : 
            for i in range(0,len(self.data),1) : 
                r = self.getRowByIndex(i)
                if r['name'] == 'initial' :
                    d.append(0)
                if r['name'] != 'initial' : 
                    d.append(self.getRowByIndex(i)['E'])                 
            d[0] = d[1]
        else : 
            print "Common.getColumn does not exist for ", colName
            return _np.array([])
        return _np.array(d)

    def containsEnergyVariation(self) : 
        '''Method to determine if the energy is constant in the lattice
        Required if there is 
        1) RfCavities
        ''' 
        for i in range(0,len(self.data),1) : 
            if self.type[i] == 'LCAV' : 
                return True
        return False 

    def getApertures(self, raw = True) : 
        aper = [] 

        # see if there is any aperture information 
        for i in range(0,len(self.data),1) :             
            aper.append(float(self.data[i][self.keys['drif']['aper']]))
        if sum(aper)== 0 :
            aper = []
            for i in range(0,len(self.data),1) :             
                aper.append(0.1)            
            return aper

        if raw : 
            for i in range(0,len(self.data),1) :             
                aper.append(self.data[i][self.keys['drif']['aper']])
            return aper
        else : 
            # find first non-zero aperture 
            for i in range(0,len(self.data),1) : 
                a = float(self.data[i][self.keys['drif']['aper']])
                if a != 0 : 
                    firstA = a
                    break 

            lastA = firstA 
            for i in range(0,len(self.data),1) : 
                a = float(self.data[i][self.keys['drif']['aper']])
                if a == 0 : 
                    a = lastA
                else : 
                    lastA = a 
                aper.append(a)

            return aper
                    
    def makeLocationList(self, elementNames = []) : 
        pass

###############################################################################################################
class Twiss(General) : 
    '''Twiss data structure
    data : numpy array of data 
    keys : key to data''' 

    keys = {'alfx':0, 'betx':1, 'mux':2, 'dx':3, 'dpx':4,
            'alfy':5, 'bety':6, 'muy':7, 'dy':8, 'dpy':9,
            'x':10,'px':11,'y':12,'py':13,'suml':14}

    def __init__(self) : 
        General.__init__(self)

    def plotBeta(self) : 
        _pl.figure(1)
        self.plotXY('suml','betx')
        self.plotXY('suml','bety')

    def plotEta(self) :
        _pl.figure(2)
        self.plotXY('suml','dx');
        self.plotXY('suml','dy');
        
    def plotEtaPrime(self) : 
        _pl.figure(3)
        self.plotXY('suml','dpx');
        self.plotXY('suml','dpy');        

    def plotAlf(self) : 
        _pl.figure(3)
        self.plotXY('suml','alfx');
        self.plotXY('suml','alfy');                

    def plotMu(self) : 
        _pl.figure(3)
        self.plotXY('suml','mux');
        self.plotXY('suml','muy');                

    def nameFromNearestS(self,s) :
        suml = self.getColumn('suml')
        for i in range(0,len(suml)-1) :
            if s > suml[i-1] and s < suml[i] :
                return self.name[i]
        return "Not found"

###############################################################################################################
class Survey(General) :
    '''Survey data structure
    data : numpy array of data 
    keys : key to data''' 
    keys = {'x':0,'y':1,'z':2,'suml':3,
            'theta':4,'phi':5,'psi':6}    

    def __init__(self) : 
        General.__init__(self)

###############################################################################################################
class Rmat(General) : 
    '''Rmatrix data structure
    data : numpy array of data 
    keys : key to data''' 

    keys = {'r11':0 ,'r12':1 ,'r13':2 ,'r14':3 ,'r15':4 ,'r16':5 ,
            'r21':6 ,'r22':7 ,'r23':8 ,'r24':9 ,'r25':10,'r26':11,
            'r31':12,'r32':13,'r33':14,'r34':15,'r35':16,'r36':17,
            'r41':18,'r42':19,'r43':20,'r44':21,'r45':22,'r46':23,
            'r51':24,'r52':25,'r53':26,'r54':27,'r55':28,'r56':29,
            'r61':30,'r62':31,'r63':32,'r64':33,'r65':34,'r66':35,
            'suml':36}    

    def __init__(self) : 
        General.__init__(self)

    def getData(self, index) :
        return self.data[index]

###############################################################################################################
class Chrom(General) : 
    '''Chromaticity data structure
    data : numpy array of data 
    keys : key to data''' 

    def __init__(self) :
        General.__init__(self)

    def getData(self, index) :
        return self.data[index]
     
###############################################################################################################           
class Envelope(General) : 
    '''Beam envelope data structure
    data : numpy array of data 
    keys : key to data'''
 
    keys = {'s11':0 ,'s12':1 ,'s13':2 ,'s14':3 ,'s15':4 ,'s16':5 ,
            's21':6 ,'s22':7 ,'s23':8 ,'s24':9 ,'s25':10,'s26':11,
            's31':12,'s32':13,'s33':14,'s34':15,'s35':16,'s36':17,
            's41':18,'s42':19,'s43':20,'s44':21,'s45':22,'s46':23,
            's51':24,'s52':25,'s53':26,'s54':27,'s55':28,'s56':29,
            's61':30,'s62':31,'s63':32,'s64':33,'s65':34,'s66':35,
            'suml':36}    

    def __init__(self) :
        General.__init__(self) 

    def getData(self,index) :
        return self.data[index]

###############################################################################################################
class OutputReader :

    '''Class to load different Mad8 output files
    Usage : 
    o = Mad8.OutputReader()
    [c, s] = o.readFile('./survey.tape','survey')    
    [c, r] = o.readFile('./rmat.tape','rmat')
    [c, t] = o.readFile('./twiss.tape','twiss')
    [c, c] = o.readFile('./chrom.tape','chrom')
    [c, e] = o.readFile('./envelope.tape','envel')

    c : Common data
    r : Rmat object 
    t : Twiss object 
    c : Chrom object 
    e : Envelope object
    '''

    def __init__(self) : 
        pass
#        import fortranformat as ff

    def readFile(self, fileName = '', type = 'twiss') :
        '''read mad8 output file
        '''
        
        self.fileName = fileName
        self.type     = type.lower() 

        if self.type == 'twiss' :
            r = self.readTwissFile()
        elif self.type == 'rmat' :
            r = self.readRmatFile()
        elif self.type == 'survey' : 
            r = self.readSurveyFile()
        elif self.type == 'chrom' :
            r = self.readChromFile() 
        elif self.type == 'envel' : 
            r = self.readEnvelopeFile()
        else : 
            print 'Mad8.OutputReader.readFile> Unknown file'
            return None 
        return r 

    def readTwissFile(self, f= None) :
        if f == None :
            f = open(self.fileName,'r')
        
        ffhr1 = _ff.FortranRecordReader('(5A8,I8,L8,I8)')
        ffhr2 = _ff.FortranRecordReader('(A80)')

#        print '"'+l+'"'                
        # read header
        h1 = ffhr1.read(f.readline())
        h2 = ffhr2.read(f.readline())
        nrec = h1[7]

        print 'Mad8.readTwissFile > nrec='+str(nrec)
        
        ffe1 = _ff.FortranRecordReader('(A4,A16,F12.6,4E16.9,A19,E16.9)')
        ffe2 = _ff.FortranRecordReader('(5E16.9)')

        common  = Common() 
        twiss   = Twiss()

        # loop over entries
        for i in range(0,nrec,1) :
            l1 = ffe1.read(f.readline())
            l2 = ffe2.read(f.readline())
            l3 = ffe2.read(f.readline())
            l4 = ffe2.read(f.readline())
            l5 = ffe2.read(f.readline())  
            common.addElement(l1[0],l1[1],l1[2:6]+l2+[l1[6],l1[7],l1[8]])          
            twiss.addElement(l1[0],l1[1].strip(),l3+l4+l5)
        common.makeArray()
        twiss.makeArray()

        # minimum and maximum s
        twiss.smin = twiss.getColumn('suml')[0]
        twiss.smax = twiss.getColumn('suml')[-1]

        # number of elements 
        common.nele = len(common.name)
        twiss.nele  = len(twiss.getColumn('suml'))

        return [common,twiss]

    def readRmatFile(self, f=None) :
        if f == None :
            f = open(self.fileName,'r')

        # Standard header definition 
        ffhr1 = _ff.FortranRecordReader('(5A8,I8,L8,I8)')
        ffhr2 = _ff.FortranRecordReader('(A80)')

        # read header
        h1 = ffhr1.read(f.readline())
        h2 = ffhr2.read(f.readline())
        # number of records
        nrec = h1[7]
        print 'Mad8.readRmatFile  > nrec='+str(nrec)

        ffe1 = _ff.FortranRecordReader('(A4,A16,F12.6,4E16.9,A19,E16.9)')
        ffe2 = _ff.FortranRecordReader('(5E16.9)')
        ffe3 = _ff.FortranRecordReader('(6E16.9)')
        ffe4 = _ff.FortranRecordReader('(7E16.9)')

        common  = Common() 
        rmat    = Rmat() 

        for i in range(0,nrec,1) :
            l1 = ffe1.read(f.readline())
            l2 = ffe2.read(f.readline())
            l3 = ffe3.read(f.readline())
            l4 = ffe3.read(f.readline())
            l5 = ffe3.read(f.readline())
            l6 = ffe3.read(f.readline())
            l7 = ffe3.read(f.readline())
            l8 = ffe4.read(f.readline())
            common.addElement(l1[0],l1[1],l1[2:6]+l2+[l1[6],l1[7],l1[8]])          
            rmat.addElement(l1[0],l1[1].strip(),l3+l4+l5+l6+l7+l8)
        f.close()    

        common.makeArray()
        rmat.makeArray()

        # minimum and maximum s
        rmat.smin = rmat.getColumn('suml')[0]
        rmat.smax = rmat.getColumn('suml')[-1]

        # number of elements 
        common.nele = len(common.name)
        rmat.nele  = len(rmat.getColumn('suml'))

        return [common,rmat]

    def readChromFile(self, f=None ) :
        if f==None : 
            f = open(self.fileName,'r') 

        # First read twiss 
        [c,twiss] = self.readTwissFile(f)
        
        # 3 random lines at end of chrom tape
        f.readline()
        f.readline()
        f.readline()

        # read header
        ffhr1 = _ff.FortranRecordReader('(5A8,I8,L8,I8)')
        ffhr2 = _ff.FortranRecordReader('(A80)')
        h1 = ffhr1.read(f.readline())
        h2 = ffhr2.read(f.readline())
        nrec = h1[7]

        print 'Mad8.readChromFile > nrec='+str(nrec)

        ffe1 = _ff.FortranRecordReader('(A4,A16,F12.6,4E16.9,A19,E16.9)')
        ffe2 = _ff.FortranRecordReader('(5E16.9)')
        ffe3 = _ff.FortranRecordReader('(6E16.9)')
        
        common = Common() 
        chrom  = Chrom() 

        for i in range(0,nrec,1) : 
            l1 = ffe1.read(f.readline())
            l2 = ffe2.read(f.readline())
            l3 = ffe3.read(f.readline())
            l4 = ffe3.read(f.readline())
            l5 = ffe3.read(f.readline())

            common.addElement(l1[0],l1[1],l1[2:6]+l2+[l1[6],l1[7],l1[8]])           
            chrom.addElement(l1[0],l1[1].strip(),l3+l4+l5)
            
        common.makeArray()
        chrom.makeArray()

        # minimum and maximum s
        chrom.smin = chrom.getColumn('suml')[0]
        chrom.smax = chrom.getColumn('suml')[-1]

        # number of elements 
        common.nele = len(common.name)
        chrom.nele  = len(chrom.getColumn('suml'))

        return [common,twiss,chrom]
    
    def readEnvelopeFile(self, f=None) : 
        if f==None : 
            f = open (self.fileName,'r') 

        ffhr1 = _ff.FortranRecordReader('(5A8,I8,L8,I8)')
        ffhr2 = _ff.FortranRecordReader('(A80)')

        # read header
        h1 = ffhr1.read(f.readline())
        h2 = ffhr2.read(f.readline())
        nrec = h1[7]

        print 'Mad8.readEnvelopeFile > nrec='+str(nrec)

        ffe1 = _ff.FortranRecordReader('(A4,A16,F12.6,4E16.9,A19,E16.9)')
        ffe2 = _ff.FortranRecordReader('(5E16.9)')
        ffe3 = _ff.FortranRecordReader('(6E16.9)')
        ffe4 = _ff.FortranRecordReader('(7E16.9)')
        
        common = Common()
        envel  = Envelope() 

        for i in range(0,nrec,1) : 
            l1 = ffe1.read(f.readline())
            l2 = ffe2.read(f.readline())
            l3 = ffe3.read(f.readline())
            l4 = ffe3.read(f.readline())
            l5 = ffe3.read(f.readline())
            l6 = ffe3.read(f.readline())
            l7 = ffe3.read(f.readline())
            l8 = ffe4.read(f.readline())

            common.addElement(l1[0],l1[1],l1[2:6]+l2+[l1[6],l1[7],l1[8]])          
            envel.addElement(l1[0],l1[1].strip(),l3+l4+l5+l6+l7+l8)
        f.close()

        common.makeArray()
        envel.makeArray()

        # minimum and maximum s
        envel.smin = envel.getColumn('suml')[0]
        envel.smax = envel.getColumn('suml')[-1]

        # number of elements 
        common.nele = len(common.name)
        envel.nele  = len(envel.getColumn('suml'))

        return [common,envel] 

    def readSurveyFile(self) :
        f = open(self.fileName,'r') 
        
        # Standard header definition 
        ffhr1 = _ff.FortranRecordReader('(5A8,I8,L8,I8)')
        ffhr2 = _ff.FortranRecordReader('(A80)')

        # read header
        h1 = ffhr1.read(f.readline())
        h2 = ffhr2.read(f.readline())
        # number of records
        nrec = h1[7]

        print 'Mad8.readSurveyFile> nrec='+str(nrec)
        ffe1 = _ff.FortranRecordReader('(A4,A16,F12.6,4E16.9,A19,E16.9)')
        ffe2 = _ff.FortranRecordReader('(5E16.9)')
        ffe3 = _ff.FortranRecordReader('(4E16.9)')
        ffe4 = _ff.FortranRecordReader('(3E16.9)')
        
        common  = Common() 
        survey  = Survey() 
                
        for i in range(0,nrec,1) :
            l1 = ffe1.read(f.readline())
            l2 = ffe2.read(f.readline())
            l3 = ffe3.read(f.readline())
            l4 = ffe4.read(f.readline())
            common.addElement(l1[0],l1[1],l1[2:6]+l2+[l1[6],l1[7],l1[8]])          
            survey.addElement(l1[0],l1[1].strip(),l3+l4)            

        f.close()
        
        common.makeArray()
        survey.makeArray()

        return [common,survey]
    
###############################################################################################################
class Mad8 :
    def __init__(self, filename) : 
        self.particle = ''
        self.subroutines = []
        self.readFile(filename) 

    def readFile(self,filename) :
        f = open(filename)
        self.file = []

        # load file
        for l in f : 
            ls = l.strip()
            if len(ls) > 0 :
                self.file.append(l.strip())
        f.close()

        # determine particle
        for l in self.file : 
            if l.find("PARTICLE") != -1:
                if l.find("ELECTRON") != -1 : 
                    self.particle = "ELECTRON"
                elif l.find("POSITRON") != -1 :
                    self.particle = "POSITRON"

        # determine subroutines
        for l in self.file :
            if l.find("SUBROUTINE") != -1: 
                sl = l.split()
                if len(sl) > 1 :                     
                    self.subroutines.append(sl[0])

###############################################################################################################
class EchoValue :
    def __init__(self,echoFileName) : 
        self.echoFileName = echoFileName
        self.valueDict = {}
        
    def loadValues(self) : 
        f = open(self.echoFileName) 

        for l in f : 
            if l.find("Value") != -1 :
                sl = l.split()
                k  = sl[3].strip('"')
                v  = float(sl[5])
                self.valueDict[k] = v

###############################################################################################################
class Saveline :
    def __init__(self, fileName, lineName = 'EBDS') : 
        self.fileName = fileName
        self.lineName = lineName

        self.dictFile            = []
        self.expandedLine        = []
        self.dictFileRenamed     = []
        self.expandedLineRenamed = []
        self.readFile(self.fileName)
        self.parseFile()
        
    def readFile(self,fileName) : 

        f = open(fileName) 
        self.file = [] 
        for l in f : 
            self.file.append(l)
        f.close()

        # tidy file
        self.file = _Input.tidy(self.file)

        # remove comments
        self.file = _Input.removeComments(self.file)

        # remove continuation symbols
        self.file = _Input.removeContinuationSymbols(self.file)
                
    def parseFile(self) :
        for l in self.file : 
            d = _Input.decodeFileLine(l)
            self.dictFile.append(d)
            
    def expandLine(self) : 
        line = self.findNamedDict(self.lineName)['LINE']
        self.expandedLine = []
        for l in line : 
            self.expandedLine = self.expandedLine + self.findNamedDict(l)['LINE']

    def findNamedIndex(self,name) :
        for i in range(0,len(self.dictFile)) :
            if self.dictFile[i]['name'] == name : 
                return i

        return -1

    def findNamedDict(self, name) : 
        idx = self.findNamedIndex(name)
        if idx != -1 : 
            return self.dictFile[idx]
        else :
            return dict()

    def findRenamedNamedIndex(self,name) :
        for i in range(0,len(self.dictFileRenamed)) :
            if self.dictFileRenamed[i]['name'] == name : 
                return i

        return -1

    def findRenamedNamedDict(self, name) : 
        idx = self.findRenamedNamedIndex(name)
        if idx != -1 : 
            return self.dictFileRenamed[idx]
        else :
            return dict()

    def removeReplacements(self) : 
        if len(self.expandedLine) == 0 :
            self.expandLine()

    def removeDuplicates(self) : 
        if len(self.expandedLine) == 0 :
            self.expandLine()

        self.nameCount = {}
    

        # find template elements (only need to write one)
        self.templates = []
        for i in range(0,len(self.dictFile)) : 
            d   = _copy.deepcopy(self.findNamedDict(self.dictFile[i]['name']))
            idx = self.findNamedIndex(d['type']) 
            if idx != -1 :                 
                try : 
                    self.templates.index(self.dictFile[idx]['name'])
                except: 
                    self.dictFileRenamed.append(self.dictFile[idx])
                    self.templates.append(self.dictFile[idx]['name'])
                        
        # loop over elements in beamline and append name 
        for i in range(0,len(self.expandedLine)) :
            oldName = self.expandedLine[i]
            d = _copy.deepcopy(self.findNamedDict(oldName))
            
            try : 
                self.nameCount[self.expandedLine[i]] = self.nameCount[self.expandedLine[i]]+1 
            except KeyError :
                self.nameCount[self.expandedLine[i]] = 0            
            
            newName = d['name']+'_'+str(self.nameCount[d['name']])
            
            d['name'] = newName
            self.dictFileRenamed.append(d)
            self.expandedLineRenamed.append(newName)
            
    def makeSubLines(self) :
        
        outputLines = []
        # split elements into 100 element chunks
        self.splitLine = [self.expandedLineRenamed[i:i+100] for i in range(0, len(self.expandedLineRenamed), 100)]
        
        # loop over split lines
        self.subLines = [] 
        for i in range(0,len(self.splitLine)) : 
            line = self.splitLine[i]
            # loop over elements in line
            lineName = 'L'+'%06i' % i
            l = lineName+': LINE = ('
            self.subLines.append(lineName)
            for le in line : 
                l = l + le
                if line.index(le) < len(line)-1 : 
                    l = l +', '
            l = l+')'
            outputLines.append(l)

        # geneate final line 
        l = self.lineName+': LINE = ('
        for sl in self.subLines : 
            l = l + sl 
            if self.subLines.index(sl) < len(self.subLines)-1 : 
                l = l +', '
            
        l = l+')'
        outputLines.append(l)

        return outputLines

    def writeRenamed(self,filename) : 
        # open file
        f = open(filename,'w')

        # write lines 
        lines = self.makeSubLines();
        for l in lines : 
            writeContinuation(f,l+'\n')

        # write components
        for i in range(0,len(self.dictFileRenamed)) :
            l = self.dictFileRenamed[i]['name']+': '+self.dictFileRenamed[i]['type']

            for k in self.dictFileRenamed[i].keys() :
                if k != 'name' and k != 'type' :
                    l = l+', '+k+'='+str(self.dictFileRenamed[i][k])
            l = l + '\n'
            writeContinuation(f,l)

        # write rturn 
        f.write('RETURN\n')

        # close file
        f.close()

def writeContinuation(f, l) : 

    iContinuation = 0

    # loop over all characters in string 
    for i in range(0,len(l)) : 
        r = _np.mod(i,60) # 60 characters before searching for a space

        if r>40 and l[i]==' ' and i > iContinuation+30: 
            iContinuation = i
            f.write('&\n')

        f.write(l[i])


class Track : 
    def __init__(self,folderpath, filemapname, twissname) :
        self.folderpath  = folderpath
        self.filemapname = filemapname
        self.twissname   = twissname
        self.trackdata = defaultdict(list)

        if not (self.folderpath.endswith("/")):
            self.folderpath = self.folderpath+"/"
        
        cwd = _os.getcwd()
        print "pymad8.Output.Track >> Initialised in directory:"
        if (self.folderpath.startswith("/")):
            print self.folderpath
        else:
            print cwd+"/"+self.folderpath

        print "pymad8.Output.Track >> Filemap file:"
        if (self.filemapname.startswith("/")):
            print self.filemapname
        else:
            print cwd+"/"+self.filemapname

        print "pymad8.Output.Track >> Twiss file:"
        if (self.twissname.startswith("/")):
            print self.twissname
        else:
            print cwd+"/"+self.twissname

              
    def readDir(self) : 
        """
        Loop over all mad8 track output files in the target directory and 
        build a dictionary of the data. File map is used to match data
        from track files to observation plane in the twiss file.
        """

        print "pymad8.Output.Track >> Loading twiss file"
        if(_os.path.isfile(self.twissname)):
            reader           = _pymad8.Output.OutputReader()
            [common , twiss] = reader.readFile(self.twissname, 'twiss')
        else:
            print "No such file:", self.twissname
            print "Terminating.."
            _sys.exit(1)
            
        print "pymad8.Output.Track >> Loading element map"
        if(_os.path.isfile(self.filemapname)):
            with open(self.filemapname, 'r') as filemap:
                fmap={line.split()[1] : int(line.split()[0]) for line in filemap.readlines()}
        else:
            print "No such file:", self.filemapname
            print "Terminating.."
            _sys.exit(1)

        print "pymad8.Output.Track >> Loading track files.."
        for fn in _os.listdir(self.folderpath):
            if _os.path.isfile(self.folderpath+fn):
                #print "Loading file ", fn
                data = _np.loadtxt(self.folderpath+fn, skiprows=51, unpack=True)
                self.trackdata[fn].append(data)

        #Get the s p_ositions from the twiss file and match them to the data using the filemap
        for name in self.trackdata.keys():
            idx  = fmap[name]
            S    = twiss.getRowByIndex(idx)["suml"]
            self.trackdata[name]=[self.trackdata[name],S]

        print "pymad8.Output.Track >> Done"
            

    def appendDir(self, folderpath):
        """
        Loop over all mad8 track output files in the target directory and 
        append the data to the existing data structure.  
        """
        if not (folderpath.endswith("/")):
            folderpath = folderpath+"/"
        
        cwd = _os.getcwd()
        print "pymad8.Output.Track >> Appending directory:"
        if (folderpath.startswith("/")):
            print self.folderpath
        else:
            print cwd+folderpath

        "pymad8.Output.Track >> Loading files...:"
        for fn in _os.listdir(folderpath):
            if _os.path.isfile(folderpath+fn):
                if(fn in self.trackdata.keys()):
                    data = _np.loadtxt(folderpath+fn, skiprows=51, unpack=True)
                    self.trackdata[fn][0][0] = _np.concatenate((self.trackdata[fn][0][0], data), axis=1)
                else:
                    print "pymad8.Output.Track >>Sampler name not found in orignal set, skip: ", fn 



import Output as _Output

import pylab as _pl
import numpy as _np
import matplotlib as _matplotlib
import matplotlib.pyplot as _plt
import matplotlib.patches as _patches

class _My_Axes(_matplotlib.axes.Axes):
    """
    Inherit matplotlib.axes.Axes but override pan action for mouse.
    Only allow horizontal panning - useful for lattice axes.
    """
    name = "_My_Axes"
    def drag_pan(self, button, key, x, y):
        _matplotlib.axes.Axes.drag_pan(self, button, 'x', x, y) # pretend key=='x'

_matplotlib.projections.register_projection(_My_Axes)

def AddMachineLatticeToFigure(figure, mad8opt, tightLayout=True):
    """
    Add a diagram above the current graph in the figure that represents the
    accelerator based on a madx twiss file in tfs format.

    Note you can use matplotlib's gcf() 'get current figure' as an argument.

    >>> pymadx.Plot.AddMachineLatticeToFigure(gcf(), 'afile.tfs')

    """    

    axs = figure.get_axes() # get existing graph

    axoptics = figure.get_axes()[0]
    _AdjustExistingAxes(figure, tightLayout=tightLayout)
    axmachine = _PrepareMachineAxes(figure)
    
    _DrawMachineLattice(axmachine,mad8opt)

    # put callbacks for linked scrolling 
    def MachineXlim(ax):
        axmachine.set_autoscale_on(False)
        axoptics.set_xlim(axmachine.get_xlim())

    def Click(a):
        if a.button == 3:
            try:
                print a.xdata,mad8opt['twiss'].nameFromNearestS(a.xdata)
            except ValueError:
                pass # don't complain if the S is out of bounds        

    MachineXlim(axmachine)
    axmachine.callbacks.connect('xlim_changed', MachineXlim)
    figure.canvas.mpl_connect('button_press_event', Click)

def _PrepareMachineAxes(figure):
    # create new machine axis with proportions 6 : 1
    axmachine = figure.add_subplot(911, projection="_My_Axes")
    axmachine.set_facecolor('none') # make background transparent to allow scientific notation
    _SetMachineAxesStyle(axmachine)
    return axmachine
    
def _AdjustExistingAxes(figure, fraction=0.9, tightLayout=True):
    """
    Fraction is fraction of height all subplots will be after adjustment.
    Default is 0.9 for 90% of height. 
    """
    # we have to set tight layout before adjustment otherwise if called
    # later it will cause an overlap with the machine diagram
    if (tightLayout):
        _plt.tight_layout()

    axs = figure.get_axes()

    for ax in axs:
        bbox = ax.get_position()
        bbox.y0 = bbox.y0 * fraction
        bbox.y1 = bbox.y1 * fraction
        ax.set_position(bbox)

def _SetMachineAxesStyle(ax):
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.spines['right'].set_visible(False)

def _DrawMachineLattice(axesinstance, mad8opt):
    ax = axesinstance
    m8 = mad8opt
    
    def DrawBend(e,color='b',alpha=1.0):
        br = _patches.Rectangle((e['suml']-e['l'],-0.1),e['l'],0.2,color=color,alpha=alpha)        
        ax.add_patch(br)
    def DrawQuad(e,color='r',alpha=1.0):
        if e['k1'] > 0 :
            qr = _patches.Rectangle((e['suml']-e['l'],0),e['l'],0.2,color=color,alpha=alpha)
        elif e['k1'] < 0:
            qr = _patches.Rectangle((e['suml']-e['l'],-0.2),e['l'],0.2,color=color,alpha=alpha)
        else:
            #quadrupole off
            qr = _patches.Rectangle((e['suml']-e['l'],-0.1),e['l'],0.2,color='#B2B2B2',alpha=0.5) #a nice grey in hex
        ax.add_patch(qr)
    def DrawHex(e,color,alpha=1.0):
        s = e['suml']-e['l']
        l = e['l']
        edges = _np.array([[s,-0.1],[s,0.1],[s+l/2.,0.13],[s+l,0.1],[s+l,-0.1],[s+l/2.,-0.13]])
        sr = _patches.Polygon(edges,color=color,fill=True,alpha=alpha)
        ax.add_patch(sr)
    def DrawRect(e,color,alpha=1.0):
        rect = _patches.Rectangle((e['suml']-e['l'],-0.1),e['l'],0.2,color=color,alpha=alpha)
        ax.add_patch(rect)
    def DrawLine(e,color,alpha=1.0):
        ax.plot([e['suml']-e['l'],e['l']-e['l']],[-0.2,0.2],'-',color=color,alpha=alpha)

    ax.plot([m8['twiss'].smin,m8['twiss'].smax],[0,0],'k-',lw=1)
    ax.set_ylim(-0.2,0.2)

    # loop over elements 
    for i in range(0,m8['comm'].nele) : 
        c = m8['comm']
        t = m8['twiss']

        # need to add s to element
        e = c.getRowByIndex(i)
        e['suml'] = t.getRowByIndex(i)['suml']

        if e['type'] == 'quad':
            DrawQuad(e, u'#d10000') #red
        elif e['type'] == 'rben':
            DrawBend(e, u'#0066cc') #blue
        elif e['type'] == 'sben':
            DrawBend(e, u'#0066cc') #blue
        elif e['type'] == 'kick':
            DrawRect(e, u'#4c33b2')
        elif e['type'] == 'hkic':
            DrawRect(e, u'#4c33b2') #purple
        elif e['type'] == 'vkic':
            DrawRect(e, u'#ba55d3')            
        elif e['type'] == 'rcol':
            DrawRect(e,'k')
        elif e['type'] == 'ecol':
            DrawRect(e,'k')
        elif e['type'] == 'sext':
            DrawHex(e, u'#ffcc00')
        elif e['type'] == 'octu':
            DrawHex(e, u'#00994c') #green
        elif e['type'] == 'lcav':
            DrawRect(e, u'#000000',0.1)
        elif e['type'] == 'sole':
            DrawRect(e, u'#000000',0.1)
        elif e['type'] == 'drif':
            pass
        elif e['type'] == 'moni':
            pass
        elif e['type'] == 'mark':
            pass
        else :
            pass
            #print 'not drawn',e['type']
        
# ============================================================================
# Below is old
# ============================================================================

class _My_Axes(_matplotlib.axes.Axes):
    """
    Inherit matplotlib.axes.Axes but override pan action for mouse.
    Only allow horizontal panning - useful for lattice axes.
    """
    name = "_My_Axes"
    def drag_pan(self, button, key, x, y):
        _matplotlib.axes.Axes.drag_pan(self, button, 'x', x, y) # pretend key=='x'

#register the new class of axes
_matplotlib.projections.register_projection(_My_Axes)

def setCallbacks(figure, axm, axplot,twiss) :
    #put callbacks for linked scrolling
    def MachineXlim(axm): 
        axm.set_autoscale_on(False)
        for ax in axplot :
            ax.set_xlim(axm.get_xlim())

    def Click(a) : 
        if a.button == 3 : 
            print 'Closest element: ',twiss.nameFromNearestS(a.xdata)

    axm.callbacks.connect('xlim_changed', MachineXlim)
    figure.canvas.mpl_connect('button_press_event', Click) 

def drawMachineLattice(mad8c, mad8t) : 
    ax = _plt.gca()
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.spines['right'].set_visible(False)

    #NOTE madx defines S as the end of the element by default
    #define temporary functions to draw individual objects
    #Not sure about mad8
    def DrawBend(e,color='b',alpha=1.0):
        br = _patches.Rectangle((e['suml']-e['l'],-0.1),e['l'],0.2,color=color,alpha=alpha)
        ax.add_patch(br)
    def DrawQuad(e,color='r',alpha=1.0):
        if e['k1'] > 0 :
            qr = _patches.Rectangle((e['suml']-e['l'],0),e['l'],0.2,color=color,alpha=alpha)
        elif e['k1'] < 0: 
            qr = _patches.Rectangle((e['suml']-e['l'],-0.2),e['l'],0.2,color=color,alpha=alpha)
        else:
            #quadrupole off
            qr = _patches.Rectangle((e['suml']-e['l'],-0.1),e['l'],0.2,color='#B2B2B2',alpha=0.5) #a nice grey in hex
        ax.add_patch(qr)
    def DrawHex(e,color,alpha=1.0):
        s = e['suml']-e['l']
        l = e['l']
        edges = _np.array([[s,-0.1],[s,0.1],[s+l/2.,0.13],[s+l,0.1],[s+l,-0.1],[s+l/2.,-0.13]])
        sr = _patches.Polygon(edges,color=color,fill=True,alpha=alpha)
        ax.add_patch(sr)
    def DrawRect(e,color,alpha=1.0):
        rect = _patches.Rectangle((e['suml']-e['l'],-0.1),e['l'],0.2,color=color,alpha=alpha)
        ax.add_patch(rect)
    def DrawLine(e,color,alpha=1.0):
        ax.plot([e['suml']-e['l'],e['suml']-e['l']],[-0.2,0.2],'-',color=color,alpha=alpha)

    # plot beam line 
    ax.plot([0,mad8t.getRowByIndex(-1)['suml']],[0,0],'k--',lw=1)
    ax.set_ylim(-0.5,0.5)

    # loop over elements and Draw on beamline
    for i in range(0,mad8c.getNElements(),1) :
        element = mad8c.getRowByIndex(i)
        element['suml'] = mad8t.getRowByIndex(i)['suml']

        kw = element['type']
        if kw == 'quad': 
            DrawQuad(element)
        elif kw == 'rbend': 
            DrawBend(element)
        elif kw == 'sben': 
            DrawBend(element)
        elif kw == 'rcol': 
            DrawRect(element,'k')
        elif kw == 'ecol': 
            DrawRect(element,'k')
        elif kw == 'sext':
            DrawHex(element,'#ffcf17') #yellow
        elif kw == 'octu':
            DrawHex(element,'g')
        elif kw == 'drif':
            pass
        elif kw == 'mult':
            element['l'] = 0.0
            DrawLine(element,'grey',alpha=0.5)
        elif kw == 'mark' :
            pass
        elif kw == '' :
            pass
        else:
            #unknown so make light in alpha
            if element['l'] > 1e-1:
                DrawRect(element,'#cccccc',alpha=0.1) #light grey
            else:
                #relatively short element - just draw a line
                DrawLine(element,'#cccccc',alpha=0.1)

    
def linearOptics(twissfile = "ebds1") : 
    r = _Output.OutputReader() 
    [c,t] = r.readFile(twissfile,"twiss")
#    [c,e] = r.readFile(name+".envelope","envel")    

    figure = _plt.figure(1)

    gs  = _plt.GridSpec(3,1,height_ratios=[1,3,3])
    ax0 = figure.add_subplot(gs[0],projection="_My_Axes")
    drawMachineLattice(c,t)    

    ax1 = _plt.subplot(gs[1]) 
    sqrtBetX = _pl.sqrt(t.getColumn("betx"))
    _plt.plot(t.getColumn("suml"),sqrtBetX,"b",label="$\\beta_{x}^{1/2}$")
    sqrtBetY = _pl.sqrt(t.getColumn("bety"))
    _plt.plot(t.getColumn("suml"),sqrtBetY,"g--",label="$\\beta_{y}^{1/2}$")
    _plt.ylabel("$\\beta_{x,y}^{1/2}$ $[{\mathrm m}]^{1/2}$")
    _plt.legend(loc=2)

    ax2 = _plt.subplot(gs[2]) 
    ax2.relim()
    ax2.autoscale_view(False,False,True)
    sqrtDisX = t.getColumn("dx")
    _plt.plot(t.getColumn("suml"),sqrtDisX,"b",label="$\\eta_{x}$")
    sqrtDisY = t.getColumn("dy")
    _plt.plot(t.getColumn("suml"),sqrtDisY,"g--",label="$\\eta_{y}$")
    _plt.xlabel("S [m]")
    _plt.ylabel("$\eta_{x,y}$ [m]")
    _plt.legend(loc=2)


    setCallbacks(figure,ax0,[ax1,ax2],t)

#    _plt.savefig(name+"_linear.pdf")
    


def phaseAdvance(twissfile = "ebds1") :
    r = _Output.OutputReader() 
    [c,t] = r.readFile(twissfile,"twiss")

    suml = t.getColumn("suml")
    muX  = t.getColumn("mux")
    muY  = t.getColumn("muy")
    
    figure = _plt.figure(1)
    gs  = _plt.GridSpec(3,1,height_ratios=[1,3,3])
    ax0 = figure.add_subplot(gs[0],projection="_My_Axes")
    drawMachineLattice(c,t)    

    ax1 = _plt.subplot(gs[1]) 
    _pl.plot(suml,muX % _pl.pi)
    _pl.ylabel("$\mu_{x}$")

    ax2 = _plt.subplot(gs[2]) 
    _pl.plot(suml,muY % _pl.pi)
    _pl.xlabel("S [m]")
    _pl.ylabel("$\mu_{y}$")

    setCallbacks(figure,ax0,[ax1,ax2],t)

#    _pl.savefig(name+"_phase.pdf")

def dispersion(twissfile = "ebds1") : 
    r = _Output.OutputReader() 
    [c,t] = r.readFile(twissfile,"twiss")

    suml = t.getColumn("suml")
    etaX  = t.getColumn("dx")
    etaY  = t.getColumn("dy")
    
    figure = _plt.figure(1)
    gs  = _plt.GridSpec(3,1,height_ratios=[1,3,3])
    ax0 = figure.add_subplot(gs[0],projection="_My_Axes")
    drawMachineLattice(c,t)     

    ax1 = _plt.subplot(gs[1]) 
    _pl.plot(suml,etaX)
    _pl.ylabel("$\eta_{x}$")

    ax2 = _plt.subplot(gs[2]) 
    _pl.plot(suml,etaY)
    _pl.xlabel("S [m]")
    _pl.ylabel("$\eta_{y}$")

    setCallbacks(figure,ax0,[ax1,ax2],t)

def dispersionPrime(twissfile = "ebds1") : 
    r = _Output.OutputReader() 
    [c,t] = r.readFile(twissfile,"twiss")

    suml = t.getColumn("suml")
    etapX  = t.getColumn("dpx")
    etapY  = t.getColumn("dpy")
    
    figure = _plt.figure(1)
    gs  = _plt.GridSpec(3,1,height_ratios=[1,3,3])
    ax0 = figure.add_subplot(gs[0],projection="_My_Axes")
    drawMachineLattice(c,t)     

    ax1 = _plt.subplot(gs[1]) 
    _pl.plot(suml,etapX)
    _pl.ylabel("$\eta_{x}\prime$")

    ax2 = _plt.subplot(gs[2]) 
    _pl.plot(suml,etapY)
    _pl.xlabel("S [m]")
    _pl.ylabel("$\eta_{y}\prime$")

    setCallbacks(figure,ax0,[ax1,ax2],t)

def apertures(twissfile="ebds1", envelfile="ebds1") : 
    # read mad8 data
    r = _Output.OutputReader() 
    [c,t] = r.readFile(twissfile,"twiss")
    [c,e] = r.readFile(envelfile,"envel")

    # calculate beam sizes
    sigmaX = _pl.sqrt(e.getColumn('s11'))
    sigmaY = _pl.sqrt(e.getColumn('s33'))
    
    # get apertures 
    aper    = _pl.array(map(float,c.getApertures(raw=False)))
    aper    = aper/aper.max()*max(sigmaX.max(),sigmaX.max())
    aperMax = aper.max()*_pl.ones(len(aper))
        
    # suml 
    suml = t.getColumn('suml')

    figure = _pl.figure(1) 
    figure.subplots_adjust(left=0.15)
    gs  = _plt.GridSpec(3,1,height_ratios=[1,3,3])

    ax0 = figure.add_subplot(gs[0],projection="_My_Axes")
    drawMachineLattice(c,t)          

    ax1 = _plt.subplot(gs[1]) 
    _plt.plot(suml,sigmaX,"b",label="$\sigma_{x}$")
    _plt.plot(suml,-sigmaX,"b")
    _plt.fill_between(suml, sigmaX, -sigmaX, color="b", alpha=0.2)
    _plt.fill_between(suml, aper, aperMax, color="k", alpha=0.2)
    _plt.fill_between(suml, -aper, -aperMax, color="k", alpha=0.2)
    _plt.ylim(-sigmaX.max(),sigmaX.max())
    _plt.ylabel("$\sigma_{x}$ [m]")


    ax2 = _plt.subplot(gs[2]) 
    _plt.plot(suml,sigmaY,"b",label="$\sigma_{y}$")
    _plt.plot(suml,-sigmaY,"b")
    _plt.fill_between(suml, sigmaY, -sigmaY, color="b", alpha=0.2)
    _plt.fill_between(suml, aper, aperMax, color="k", alpha=0.2)
    _plt.fill_between(suml, -aper, -aperMax, color="k", alpha=0.2)
    _plt.xlabel("s [m]")
    _plt.ylabel("$\sigma_{y}$ [m]")
    _plt.ylim(-sigmaX.max(),sigmaX.max())

    setCallbacks(figure,ax0,[ax1,ax2],t)

#    _plt.savefig(name+"_apertures.pdf")


def energy(twissfile = "ebds1") : 
    # read mad8 data
    r = _Output.OutputReader() 
    [c,t] = r.readFile(twissfile,"twiss")

    # get suml 
    suml = t.getColumn('suml')[1:]
    
    figure = _pl.figure(1)
    figure.subplots_adjust(left=0.15)
    gs  = _plt.GridSpec(2,1,height_ratios=[1,6])
    
    ax0 = figure.add_subplot(gs[0],projection="_My_Axes")
    drawMachineLattice(c,t)      

    ax1 = _plt.subplot(gs[1]) 
    e   = c.getColumn("E")
    _plt.plot(suml,e,"b",label="$E$")
    _plt.xlabel("$s$ [m]")
    _plt.ylabel("$E$ [GeV]")
    _plt.legend()
    
    setCallbacks(figure,ax0,[ax1],t)
    
def survey(surveyfile = "ebds1") : 
    # read mad8 data
    r = _Output.OutputReader()
    [c,s] = r.readFile(surveyfile,"survey") 
    
    # get suml
    suml = s.getColumn('suml')
    
    print suml
